<?php

/**
 * @file
 * Classes that extend Search Lucene API's abstract and base classes.
 */

/**
 * We encode everything as UTF-8, string recognized by both Zend and iconv.
 */
define('LUCENEAPI_ENCODING', 'UTF-8');


class LuceneapiQueryConverterZendFramework extends LuceneapiQueryConverter {

  /**
   * Converts a Zend_Search_Lucene_Search_Query object to a Search Lucene API
   * query array.
   */
  public function convertQuery($query_obj) {
    $query_array = array();
    $this->_convertQueryClause($query_obj, $query_array, TRUE);
    return $query_array;
  }

  /**
   * Converts an individual search clause to a Search Lucene API query array.
   */
  protected function _convertQueryClause($query_obj, &$query_array, $sign) {
    switch (TRUE) {
      case $query_obj instanceof Zend_Search_Lucene_Search_Query_Boolean:
        $query_array[] = array(
          '#type' => 'boolean',
          '#boost' => $query_obj->getBoost(),
          '#sign' => $sign,
        );
        $key = count(element_children($query_array)) - 1;

        $signs = $query_obj->getSigns();
        foreach ($query_obj->getSubqueries() as $subquery_key => $subquery) {
          $this->_convertQueryClause($subquery, $query_array[$key], $signs[$subquery_key]);
        }

        break;

      case $query_obj instanceof Zend_Search_Lucene_Search_Query_Empty:
        $query_array[] = array('#type' => 'empty');
        break;

      case $query_obj instanceof Zend_Search_Lucene_Search_Query_Preprocessing_Fuzzy:
        $query_array[] = array(
          '#type' => 'fuzzy',
          '#text' => $query_obj->getWord(),
          '#field' => $query_obj->getField(),
          '#boost' => $query_obj->getBoost(),
          '#minimum_similarity' => $query_obj->getMinimumSimilarity(),
          '#sign' => $sign,
        );
        break;

      case $query_obj instanceof Zend_Search_Lucene_Search_Query_Insignificant:
        $query_array[] = array('#type' => 'insignificant');
        break;

      case $query_obj instanceof Zend_Search_Lucene_Search_Query_Preprocessing_Phrase:
        $query_array[] = array(
          '#type' => 'phrase',
          '#text' => $query_obj->getPhrase(),
          '#field' => $query_obj->getField(),
          '#boost' => $query_obj->getBoost(),
          '#slop' => $query_obj->getSlop(),
          '#sign' => $sign,
        );
        break;

      case $query_obj instanceof Zend_Search_Lucene_Search_Query_Range:
        $query_array[] = array(
          '#type' => 'range',
          '#lower_term' => $query_obj->getLowerTerm()->text,
          '#upper_term' => $query_obj->getUpperTerm()->text,
          '#is_inclusive' => $query_obj->isInclusive(),
          '#field' => $query_obj->getField(),
          '#boost' => $query_obj->getBoost(),
          '#sign' => $sign,
        );
        break;

      case $query_obj instanceof Zend_Search_Lucene_Search_Query_Preprocessing_Term:
        $query_array[] = array(
          '#type' => 'term',
          '#text' => $query_obj->getWord(),
          '#field' => $query_obj->getField(),
          '#boost' => $query_obj->getBoost(),
          '#sign' => $sign,
        );
        break;

      case $query_obj instanceof Zend_Search_Lucene_Search_Query_MultiTerm:
        $query_array[] = array(
          '#type' => 'multiterm',
          '#boost' => $query_obj->getBoost(),
          '#sign' => $sign,
        );
        $key = count(element_children($query_array)) - 1;

        $signs = $query_obj->getSigns();
        foreach ($query_obj->getQueryTerms() as $term_key => $term) {
          $query_array[$key][] = array(
            '#text' => $term->text,
            '#field' => $term->field,
            '#sign' => $signs[$term_key],
          );
        }

        break;

      default:
        throw new Exception(t(
          'Query type %class not supported.',
          array('%class' => get_class($query_obj))
        ));
        break;
    }
  }

  /**
   * Converts a Search Lucene API query array to a Zend Framework object.
   */
  public function buildQuery(array $query_array) {
    try {
      return $this->_buildQueryClause($query_array);
    }
    catch (Exception $e) {
      drupal_set_message($e->getMessage(), 'error');
      watchdog_exception('luceneapi_zf', $e);
      return FALSE;
    }
  }

  /**
   *
   */
  public function _buildQueryClause(array $query_array) {

    if (empty($query_array['#type'])) {
      $query_array['#type'] = 'boolean';
    }

    switch ($query_array['#type']) {
      case 'keys':
        $query_obj = Zend_Search_Lucene_Search_QueryParser::parse($query_array['#text']);
        break;

      case 'boolean':
        $subqueries = $signs = array();
        foreach (element_children($query_array) as $key) {
          $subqueries[] = $this->_buildQueryClause($query_array[$key]);
          $signs[] = (isset($query_array[$key]['#sign'])) ? $query_array[$key]['#sign'] : NULL;
        }
        if (empty($subqueries)) {
          $subqueries = $signs = NULL;
        }
        $query_obj = new Zend_Search_Lucene_Search_Query_Boolean($subqueries, $signs);
        break;

      case 'empty':
        $query_obj = new Zend_Search_Lucene_Search_Query_Empty();
        break;

      case 'fuzzy':
        if (!isset($query_array['#minimum_similarity'])) {
          $query_array['#minimum_similarity'] = Zend_Search_Lucene_Search_Query_Fuzzy::DEFAULT_MIN_SIMILARITY;
        }
        $text = $this->normalizeText($query_array['#text'], $query_array['#field']);
        $query_obj = new Zend_Search_Lucene_Search_Query_Preprocessing_Fuzzy(
          $text, LUCENEAPI_ENCODING, $query_array['#field'], $query_array['#minimum_similarity']
        );
        break;

      case 'insignificant':
        $query_obj = new Zend_Search_Lucene_Search_Query_Insignificant();
        break;

      case 'multiterm':
        $terms = $signs = array();
        foreach (element_children($query_array) as $key) {
          $text = $this->normalizeText($query_array[$key]['#text'], $query_array[$key]['#field']);
          $terms[] = new Zend_Search_Lucene_Index_Term($text, $query_array[$key]['#field']);
          $signs[] = (isset($query_array[$key]['#sign'])) ? $query_array[$key]['#sign'] : NULL;
        }
        $query_obj = new Zend_Search_Lucene_Search_Query_MultiTerm($terms, $signs);
        break;

      case 'phrase':
        $text = $this->normalizeText($query_array['#text'], $query_array['#field']);
        $query_obj = new Zend_Search_Lucene_Search_Query_Preprocessing_Phrase(
          $text, LUCENEAPI_ENCODING, $query_array['#field']
        );
        break;

      case 'range':
        if (!isset($query_array['#is_inclusive'])) {
          $query_array['#is_inclusive'] = FALSE;
        }
        $lower_term = $this->normalizeText($query_array['#lower_term'], $query_array['#field']);
        $upper_term = $this->normalizeText($query_array['#upper_term'], $query_array['#field']);
        $query_obj = new Zend_Search_Lucene_Search_Query_Range(
          new Zend_Search_Lucene_Index_Term($lower_term, $query_array['#field']),
          new Zend_Search_Lucene_Index_Term($upper_term, $query_array['#field']),
          $query_array['#is_inclusive']
        );
        break;

      case 'term':
        $text = $this->normalizeText($query_array['#text'], $query_array['#field']);
        $query_obj = new Zend_Search_Lucene_Search_Query_Preprocessing_Term(
          $text, LUCENEAPI_ENCODING, $query_array['#field']
        );
        break;

      case 'wildcard':
        $pattern = $this->normalizeText($query_array['#pattern'], $query_array['#field']);
        $query_obj = new Zend_Search_Lucene_Search_Query_Term(
          new Zend_Search_Lucene_Index_Term($pattern, $query_array['#field'])
        );
        break;

      default:
        throw new Exception(t(
          'Query type %type not valid.',
          array('%type' => $query_array['#type'])
        ));
    }

    if (isset($query_array['#boost'])) {
      $query_obj->setBoost($query_array['#boost']);
    }

    return $query_obj;
  }
}

/**
 *
 *
 */
class LuceneapiHitZendFramework Extends LuceneapiHit {

  /**
   *
   */
  public function setId($hit) {
    $this->id = $hit->id;
  }

  /**
   *
   */
  public function setScore($hit) {
    $this->score = $hit->score;
  }

  /**
   *
   */
  public function getFieldValue($field) {
    try {
      return $this->_hit->$field;
    }
    catch (Exception $e) {
      return '';
    }
  }
}

/**
 * Gets stopwords from settings.
 *
 * @param $text
 *   A string containing the raw setting text.
 * @return
 *   An array of stopwords.
 */
function luceneapi_zf_parse_stopwords($text) {
  $stopwords = array();
  $continue = TRUE;
  $byte_position = 0;
  do {
    if (preg_match('/[\p{L}\p{N}]+/u', $text, $match, PREG_OFFSET_CAPTURE, $byte_position)) {
      $stopwords[] = drupal_strtolower($match[0][0]);
      $byte_position = $match[0][1] + strlen($match[0][0]);
    }
    else {
      $continue = FALSE;
    }
  }
  while ($continue);
  return $stopwords;
}
