<?php

/**
 * @file
 * Abstract classes and base classes.
 */

/**
 * Extended by classes that load settings via the CTools export component.
 */
abstract class LuceneapiSettings {

  /**
   * An object containing the settings loaded via CTools.
   *
   * @var stdClass
   */
  protected $settings;

  /**
   * A registry of plugin definitions grouped by type.
   *
   * @var array
   */
  protected $plugins = array();

  /**
   * Loads settings via the CTools API function, throws an Exception if the
   * settings could not be loaded.
   *
   * @param $name
   *   A string containing the name of the configuration.
   *
   * @throws Exception
   * @see ctools_export_crud_load()
   */
  public function __construct($name) {
    ctools_include('export');
    if (!$this->settings = ctools_export_crud_load($this->getTable(), $name)) {
      throw new Exception(t(
        'Setting %name not defined in table %table.',
        array('%name' => $name, '%table' => $this->getTable())
      ));
    }
  }

  /**
   * Returns the table that stores the configuration.
   *
   * @return
   *   A string containing the table.
   */
  abstract public function getTable();

  /**
   * Gets the machine readable name of the configuration.
   *
   * @return
   *   A string containing the machine readable name of the configuration.
   */
  public function getName() {
    return $this->settings->name;
  }

  /**
   * Gets the human readable name of the configuration.
   *
   * @return
   *   A string containing the machine readable name of the configuration.
   */
  public function getLabel() {
    return $this->settings->label;
  }

  /**
   * Gets the description of the configuration.
   *
   * @return
   *   A string containing the machine readable name of the configuration.
   */
  public function getDescription() {
    return $this->settings->description;
  }

  /**
   * Retrieves the settings loaded via the CTools API.
   *
   * @return
   *   An object containing the settings.
   */
  public function getSettings() {
    return $this->settings;
  }

  /**
   * Uses the CTools to convert the settings to something that is exportable.
   *
   * @return
   *   The export value.
   */
  public function getExport() {
    return ctools_export_crud_export($this->getTable(), $this->settings);
  }

  /**
   * Safely retrieves the value of a setting.
   *
   * @param $settings
   *   A string containing the name of the setting.
   *
   * @return
   *   The settings value, NULL if the setting is not stored.
   */
  public function getSetting($setting) {
    if (isset($this->settings->settings[$setting])) {
      return $this->settings->settings[$setting];
    }
  }

  /**
   * Registers a plugin.
   *
   * @param $type
   *   The type of plugin.
   * @param $id
   *   The id of the specific plugin to register.
   *
   * @return
   *   An instance of this class.
   */
  public function registerPlugin($type, $id) {
    ctools_get_plugins('luceneapi', $type, $id);
    $this->plugins[$type][$id] = ctools_get_plugins('luceneapi', $type, $id);
    return $this;
  }

  /**
   * Unregisters a plugin.
   *
   * @param $type
   *   The type of plugin.
   * @param $id
   *   The id of the specific plugin to unregister.
   *
   * @return
   *   An instance of this class.
   */
  public function unregisterPlugin($type, $id) {
    unset($this->plugins[$type][$id]);
    return $this;
  }

  /**
   * Returns registered plugins.
   *
   * @param $type
   *   The type of plugin.
   *
   * @return
   *   An array containing the registered plugins.
   */
  public function getPlugins($type = NULL) {
    if (NULL !== $type) {
      return isset($this->plugins[$type]) ? $this->plugins[$type] : array();
    }
    else {
      return $this->plugins;
    }
  }

  /**
   * Returns a registered plugin definition.
   *
   * @param $type
   *   The type of plugin.
   * @param $id
   *   The id of the specific plugin being returned.
   *
   * @return
   *   An array containing the plugin.
   */
  public function getPlugin($type, $id) {
    if (!empty($this->plugins[$type][$id])) {
      return $this->plugins[$type][$id];
    }
    else {
      return FALSE;
    }
  }
}

/**
 * Abstract class extended by the backend that interacts with the index.
 */
class LuceneapiIndex extends LuceneapiSettings {

  /**
   * The backend plugin.
   *
   * @var LuceneapiBackend
   */
  protected $backend;

  /**
   * The indexer plugin.
   *
   * @var LuceneapiIndexer
   */
  protected $indexer;

  /**
   * Maps normalizer hooks to callback functions.
   *
   * @var array
   */
  protected $normalizers = array();

  /**
   * Loads settings, instantiates indexer and backend objects.
   */
  public function __construct($name) {
    parent::__construct($name);

    // Loads backend object.
    if (empty($this->settings->backend)) {
      throw new Exception(t(
        'Backend required for index "@name".',
        array('@name' => $this->settings->backend)
      ));
    }

    // Loads the backend plugin class.
    if ($class = luceneapi_get_plugin_class('backends', $this->settings->backend)) {
      $this->backend = new $class($this);
    }
    else {
      throw new Exception(t(
        'Error loading "@plugin" plugin.',
        array('@plugin' => $this->settings->backend)
      ));
    }

    // Loads the indexer object.
    if (empty($this->settings->indexer)) {
      throw new Exception(t(
        'Indexer required for index "@name".',
        array('@name' => $name)
      ));
    }
    elseif (!$this->indexer = luceneapi_indexer_load($this->settings->indexer)) {
      throw new Exception(t(
        'Error loading "@plugin" plugin.',
        array('@plugin' => $this->settings->indexer)
      ));
    }

    // Allows modules to alter the index.
    module_invoke_all('luceneapi_index_open', $this);

    // Instantiates normalizer plugins.
    // @todo Let's figure out how to lazy load these.
    foreach ($this->getPlugins('normalizers') as $plugin) {
      $class = ctools_plugin_get_class($plugin, 'handler');
      $instance = new $class();
      $this->normalizers[$instance->getDataType()][] = $instance;
    }
  }

  /**
   * Returns the table storing the index configurations.
   */
  public function getTable() {
    return 'luceneapi_index';
  }

  /**
   * Returns the backend.
   */
  public function getBackend() {
    return $this->backend;
  }

  /**
   * Returns the indexer.
   */
  public function getIndexer() {
    return $this->indexer;
  }

  /**
   *
   */
  public function getFields() {
    return $this->indexer->getFields();
  }

  /**
   *
   */
  public function getEnabledFields() {
    return array_intersect_key(
      $this->indexer->getFields(),
      array_filter($this->getSetting('enabled_fields'))
    );
  }

  /**
   *
   */
  public function getSortableFields() {
    return array_intersect_key(
      array_filter($this->getFields(), 'luceneapi_filter_sortable_field'),
      $this->getEnabledFields()
    );
  }

  /**
   *
   */
  public function getBiasFields() {
    return array_intersect_key(
      array_filter($this->getFields(), 'luceneapi_filter_bias_field'),
      $this->getEnabledFields()
    );
  }

  /**
   * Calls the normalize hook in the appropriate normalizer class to act on the
   * text.
   *
   * @param $text
   *
   * @param $field
   *
   * @return
   *   A string containing the normalized text.
   */
  public function normalizeText($text, array $field) {
    if (isset($this->normalizers[$field['data type']])) {
      foreach ($this->normalizers[$field['data type']] as $normalizer) {
        $text = $normalizer->execute($text, $field['data type options']);
      }
    }
    return $text;
  }

  /**
   * Returns content queued to be indexed.
   */
  public function getQueue() {
    $limit = $this->getSetting('update_limit');
    return $this->getIndexer()->getQueue($this, $limit);
  }

  /**
   * Indexes all items in the queue.
   */
  public function indexItems() {
    $vars = array('@name' => $this->getName());
    watchdog('luceneapi', 'Index update started: @name', $vars);
    $this->backend->preIndex();
    foreach ($this->getQueue() as $item) {
      try {
        $this->indexItem($item);
      }
      catch (Exception $e) {
        watchdog_exception('luceneapi', $e);
      }
    }
    $this->backend->postIndex();
    watchdog('luceneapi', 'Index update complete: @name', $vars);
  }

  /**
   * Indexes an item.
   */
  public function indexItem($item) {
    try {

      // Query to delete by node ID.
      $delete_query = array(
        '#type' => 'term',
        '#text' => $item->id,
        '#field' => 'nid',
      );

      // Invokes preBuild() hook, builds item for indexing.
      $this->indexer->preBuild($item, $this);
      $build = $this->indexer->buildItem($item, $this);

      // Instantiates document, iterates over fields and adds to document.
      $document = $this->backend->newDocument();
      foreach ($this->getEnabledFields() as $field_name => $field) {

        // Extracts text through the field's extract callback.
        $arguments = array($build, $field);
        if (!empty($field['extract arguments'])) {
          $arguments = array_merge($arguments, $field['extract arguments']);
        }
        $text = call_user_func_array($field['extract callback'], $arguments);
        if (FALSE === $text) {
          $text = (isset($field['default value'])) ? $field['default value'] : '';
        }

        // Did the extract callback returns text, or text for multiple fields?
        // An example of where multiple fields are returned is the nodeaccess
        // extract callback luceneapi_node_extract_nodeaccess().
        $index_fields = (!is_array($text)) ? array($field_name => $text) : $text;

        // Iterate over all fields to be indexed, normalizes text and adds to
        // the Lucene document.
        foreach ($index_fields as $index_field_name => $text) {
          $text = $this->normalizeText($text, $field);
          $this->backend->addField($document, $field, $index_field_name, $text);
        }
      }

      // Removes old document from index, updates new document.
      $this->backend->deleteDocumentByQuery($delete_query);
      $this->backend->addDocument($document);

    }
    catch (Exception $e) {
      try {
        watchdog_exception('luceneapi', $e);
        $this->backend->deleteDocumentByQuery($delete_query);
      }
      catch (Exception $e) {
        watchdog_exception('luceneapi', $e);
      }
    }
  }
}

/**
 * Performs searches.
 */
class LuceneapiSearcher extends LuceneapiSettings {

  /**
   * The LuceneapiIndex object associate with this searcher.
   */
  protected $index;

  /**
   *
   */
  protected $_bias;

  /**
   *
   */
  protected $_sortable;

  /**
   * The highlighter class name.
   *
   * @var string
   */
  protected $highlighterClassName;

  /**
   * Loads the index object.
   */
  public function __construct($name) {
    parent::__construct($name);

    // Instantiates the backend class.
    if (empty($this->settings->index_name)) {
      throw new Exception(t(
        'Index required for searcher "@name".',
        array('@name' => $name)
      ));
    }

    if (!$this->index = luceneapi_index_load($this->settings->index_name)) {
      throw new Exception(t(
        'Error loading index "@name".',
        array('@name' => $this->settings->index_name)
      ));
    }

    // Loads the highlighter class name.
    $plugin = $this->index->getPlugin('highlighters', $this->getSetting('highlighter'));
    $this->highlighterClassName = ctools_plugin_get_class($plugin, 'handler');
  }

  /**
   *
   */
  public function getTable() {
    return 'luceneapi_searcher';
  }

  /**
   * Returns the LuceneapiIndex object associated with this searcher.
   *
   * @return
   *   A LuceneapiIndex object.
   */
  public function getIndex() {
    return $this->index;
  }

  /**
   * Returns an instantiated highlighter plugin.
   *
   * @param $text
   *   A string containing the text to be highlighted.
   *
   * @return
   *   A LuceneapiHighlighter instance.
   *
   * @todo Error handling
   */
  public function getHighlighter($text) {
    return new $this->highlighterClassName($text, $this);
  }

  /**
   * Executes a search, returns the matches.
   *
   * @param $query
   *   An array containing the search query.
   * @return
   *   An array of search "hits".
   */
  public function search(array $query) {
    try {
      $this->index->getBackend()->setTotalRows(0);
      return $this->index->getBackend()->search($query, $this);
    }
    catch (Exception $e) {
      luceneapi_handle_exception($e);
    }
    return array();
  }

  /**
   *
   */
  public function getTotalRows() {
    return $this->index->getBackend()->getTotalRows();
  }

  /**
   * Builds the search result for the matched item.
   *
   * @param $hit
   *   A LuceneapiHit object.
   * @return
   *   An array suitable for theming.
   * @todo Error handling
   */
  public function buildResult(LuceneapiHit $hit) {
    return $this->index->getIndexer()->buildResult($hit, $this);
  }
}

/**
 *
 *
 */
abstract class LuceneapiHit {

  /**
   *
   */
  public $id;

  /**
   *
   */
  public $score;

  /**
   *
   */
  protected $_hit;

  /**
   *
   */
  public function __construct($hit) {
    $this->_hit = $hit;
    $this->setId($hit);
  }

  /**
   *
   */
  abstract public function setId($hit);

  /**
   *
   */
  abstract public function setScore($hit);

  /**
   *
   */
  abstract public function getFieldValue($field);

  /**
   *
   */
  public function __get($field) {
    return $this->getFieldValue($field);
  }
}

/**
 * Optional abstract class excended by classes that perform conversions between
 * the backend's native query API to Search Lucene API's query array structure.
 */
abstract class LuceneapiQueryConverter {

  /**
   * The LuceneapiIndex object.
   */
  protected $index;

  /**
   * Constructor, sets index object.
   */
  public function __construct(LuceneapiIndex $index) {
    $this->index = $index;
  }

  /**
   * Calls the normalize hook in the appropriate normalizer class to act on the
   * text.
   *
   * @param $text
   *
   * @param $field
   *
   * @return
   *   A string containing the normalized text.
   */
  public function normalizeText($text, $field) {
    if (!is_array($field)) {
      $field = $this->index->getIndexer()->getField($field);
    }
    return $this->index->normalizeText($text, $field);
  }

  /**
   * Converts the Search Lucene API query array into the the backends native
   * query object.
   *
   * @param $query_array
   *   A mixed value containing the backend's native query object.
   * @return
   *   An array containing the Search Lucene API query array.
   */
  abstract public function buildQuery(array $query_array);

}
